/*****************************************************************************
 *
 * This file is part of Mapnik (c++ mapping toolkit)
 *
 * Copyright (C) 2021 Artem Pavlenko
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *****************************************************************************/

#ifndef MAPNIK_UTIL_NONCOPYABLE_HPP
#define MAPNIK_UTIL_NONCOPYABLE_HPP

namespace mapnik { namespace util {

namespace non_copyable_
{

class movable
{
protected:
    constexpr movable() = default;
    ~movable() = default;
    movable( movable && ) = default;
    movable& operator=(movable && ) = default;
};

class noncopyable
{
protected:
    constexpr noncopyable() = default;
    ~noncopyable() = default;
    noncopyable( noncopyable const& ) = delete;
    noncopyable& operator=(noncopyable const& ) = delete;
};

}

using movable = non_copyable_::movable;
using noncopyable = non_copyable_::noncopyable;

}}

#endif  // MAPNIK_UTIL_NONCOPYABLE_HPP
